#!/usr/bin/env python
# $Id$
# -*- coding: utf-8 -*-

import logging
import urllib2
import xmlrpclib
import cookielib
import getpass
import logging

import kerberos as krb

class GssapiAuthError(Exception):
    """raised on error during authentication process"""

import re
RGX = re.compile('(?:.*,)*\s*Negotiate\s*([^,]*),?', re.I)

def get_negotiate_value(headers):
    for authreq in headers.getheaders('www-authenticate'):
        match = RGX.search(authreq)
        if match:
            return match.group(1)

class HTTPGssapiAuthHandler(urllib2.BaseHandler):
    """Negotiate HTTP authentication using context from GSSAPI"""

    handler_order = 400 # before Digest Auth

    def __init__(self):
        self._reset()

    def _reset(self):
        self._retried = 0
        self._context = None

    def clean_context(self):
        if self._context is not None:
            krb.authGSSClientClean(self._context)

    def http_error_401(self, req, fp, code, msg, headers):
        try:
            if self._retried > 5:
                raise urllib2.HTTPError(req.get_full_url(), 401,
                                        "negotiate auth failed", headers, None)
            self._retried += 1
            logging.debug('gssapi handler, try %s' % self._retried)
            negotiate = get_negotiate_value(headers)
            if negotiate is None:
                logging.debug('no negotiate found in a www-authenticate header')
                return None
            logging.debug('HTTPGssapiAuthHandler: negotiate 1 is %r' % negotiate)
            result, self._context = krb.authGSSClientInit("HTTP@%s" % req.get_host())
            if result < 1:
                raise GssapiAuthError("HTTPGssapiAuthHandler: init failed with %d" % result)
            result = krb.authGSSClientStep(self._context, negotiate)
            if result < 0:
                raise GssapiAuthError("HTTPGssapiAuthHandler: step 1 failed with %d" % result)
            client_response = krb.authGSSClientResponse(self._context)
            logging.debug('HTTPGssapiAuthHandler: client response is %s...' % client_response[:10])
            req.add_unredirected_header('Authorization', "Negotiate %s" % client_response)
            server_response = self.parent.open(req)
            negotiate = get_negotiate_value(server_response.info())
            if negotiate is None:
                logging.warning('HTTPGssapiAuthHandler: failed to authenticate server')
            else:
                logging.debug('HTTPGssapiAuthHandler negotiate 2: %s' % negotiate)
                result = krb.authGSSClientStep(self._context, negotiate)
                if result < 1:
                    raise GssapiAuthError("HTTPGssapiAuthHandler: step 2 failed with %d" % result)
            return server_response
        except GssapiAuthError, exc:
            logging.error(repr(exc))
        finally:
            self.clean_context()
            self._reset()

class HTTPSGssapiTransport(xmlrpclib.SafeTransport):
    """
        XMLRPC Transport that uses HTTPGssapiAuthHandler so that we can do GSSAPI authentication
    """

    def __init__(self, realm, cookiejar = cookielib.CookieJar(), username = getpass.getuser(), verbose = None, use_datetime=0):
        self._username = username
        self._realm = realm
        self.verbose = verbose
        self._use_datetime = use_datetime
        self._connection = (None, None)
        # handle cookies
        self.__cj = cookiejar

    def single_request(self, host, handler, request_body, verbose=0):
        # issue XML-RPC request

        h = self.make_connection(host)
        if verbose:
            h.set_debuglevel(1)

        try:
            self.send_request(h, handler, request_body)
            self.send_host(h, host)
            self.send_user_agent(h)
            self.send_content(h, request_body)

            response = h.getresponse(buffering=True)
            if response.status == 200:
                self.verbose = verbose

                # read response data from httpresponse, and parse it
                if response.getheader("Content-Encoding", "") == "gzip":
                    stream = GzipDecodedResponse(response)
                else:
                    stream = response
                return self.parse_response(stream)
        except Fault:
            raise
        except Exception:
            # All unexpected errors leave connection in
            # a strange state, so we clear it.
            self.close()
            raise

        #discard any response data and raise exception
        if (response.getheader("content-length", 0)):
            response.read()
        raise ProtocolError(
            host + handler,
            response.status, response.reason,
            response.msg,
            )
    
    def parse_response(self, file):
        p, u = self.getparser()

        while 1:
            data = file.read(1024)
            if not data:
                break
            if self.verbose:
                print "body:", repr(data)
            p.feed(data)

        file.close()
        p.close()

        return u.close()

    def request(self, host, handler, request_body, verbose):
        url='https://'+host+handler
        if verbose or self.verbose:
            print "ProxyTransport URL: [%s]"%url
        
        request = urllib2.Request(url)
        request.add_data(request_body)
        # Note: 'Host' and 'Content-Length' are added automatically
        request.add_header("User-Agent", self.user_agent)
        request.add_header("Content-Type", "text/xml") # Important
        
        # setup gssapi authentication
        authhandler = HTTPGssapiAuthHandler()
        ch = urllib2.HTTPCookieProcessor(self.__cj)
        opener = urllib2.build_opener(authhandler, ch)
        
        f=opener.open(request)
        return(self.parse_response(f))

if __name__ == '__main__':
    import sys
    # FOR TRAC make sure in trac.ini
    # [components] 
    # tracrpc.* = enabled 
    # Is set and XML_RPC Perms are granted to "authenticated" 
    # subject under trac->admin->General->Permissions
    #
    # debug
    logging.basicConfig(level=logging.INFO)
    gssapiTransport = HTTPSGssapiTransport("REDHAT.COM", verbose=0)
    server = xmlrpclib.ServerProxy("https://engineering.redhat.com/trac/test/login/xmlrpc", transport=gssapiTransport)
    try:
        multicall = xmlrpclib.MultiCall(server)
        for method in server.system.listMethods():
            multicall.system.methodHelp(method)

        for help in multicall():
            lines = help.splitlines()
            print lines[0]
            print '\n'.join(['  ' + x for x in lines[2:]])
            print 
    
    except xmlrpclib.ProtocolError, err:
        print "A protocol error occurred"
        print "URL: %s" % err.url
        print "HTTP/HTTPS headers: %s" % err.headers
        print "Error code: %d" % err.errcode
        print "Error message: %s" % err.errmsg

# vim:syntax=python
# vim:sw=4:softtabstop=4:expandtab
